////////////////////////////////////////////////////////////////////////////////
// auls_grapheditor.cpp
////////////////////////////////////////////////////////////////////////////////

#include <aulslib/generic.h>
#include <aulslib/memory.h>
#include <aulslib/window.h>
#include <aulslib/gdi.h>
#include <aulslib/aviutl.h>
#include <aulslib/lua.h>

#include "resource.h"

using namespace auls;

////////////////////////////////////////////////////////////////////////////////
// 萔錾
////////////////////////////////////////////////////////////////////////////////

static const LPSTR FILTER_NAME = "OtGfB^";
static const LPSTR FILTER_INFORMATION = "AulsOtGfB^ 0.01";

static const int GE_MAX_POINT = 32;
static const int GE_MAX_GRAPH = 64;
static const int GE_CLIENT_WIDTH = 512;
static const int GE_CLIENT_HEIGHT = 256;
static const int GE_POINT_SIZE = 4;

static const COLORREF GE_GRAPH_COLOR = RGB(64, 64, 64);

static const LPCSTR GE_CLASSNAME = "auls grapheditor";

struct GE_POINT {
	short x, y;
};

struct GE_GRAPHDATA {
	int num;
	struct POINT { short x, y; } pt[GE_MAX_POINT];
};

class CGeGraphData {
public:
	CStaticArray<GE_POINT, GE_MAX_POINT> pts;

	void Init(void)
	{
		pts[0].x = 0; pts[0].y = GE_CLIENT_HEIGHT - 1;
		pts[1].x = GE_CLIENT_WIDTH - 1; pts[1].y = 0;
		pts.size = 2;
	}

	void MovePoint(int idx, int x, int y)
	{
		if(idx < 0 || idx >= GE_MAX_POINT)  return;

		if(y < 0) y = 0;
		else if(y >= GE_CLIENT_HEIGHT) y = GE_CLIENT_HEIGHT-1;

		if(idx == 0) {
			x = 0;
		} else if(idx == pts.size-1) {
			x = GE_CLIENT_WIDTH-1;
		} else {
			if(x <= pts[idx-1].x) x = pts[idx-1].x+1;
			else if(x >= pts[idx+1].x) x = pts[idx+1].x-1;
		}

		pts[idx].x = x;
		pts[idx].y = y;
	}

	bool InsertPoint(int x, int y)
	{
		if(pts.IsMax()) return false;

		// ̃|CgƓWɒǉȂ悤ɂ
		for(int i = pts.size; i;) {
			i--;
			if(x == pts[i].x) x--;
			else if(x > pts[i].x) break;
		}
		if(x < 0) return false;

		// |Cgǉ
		for(int i = 0; i < (int)pts.size; i++) {
			if(x < pts[i].x) {
				GE_POINT pt = { (short)x, (short)y };
				pts.Insert(pt, i);
				break;
			}
		}
		return true;
	}

	bool ErasePoint(int idx)
	{
		if(idx > 0 && idx < (int)pts.size-1) {
			pts.Erase(idx);
			return true;
		} else {
			return false;
		}
	}

	int HitCheck(int x, int y)
	{
		for(int i = pts.size; i;) {
			i--;
			if(x >= pts[i].x - GE_POINT_SIZE && x <= pts[i].x + GE_POINT_SIZE &&
				y >= pts[i].y - GE_POINT_SIZE && y <= pts[i].y + GE_POINT_SIZE) return i;
		}
		return -1;
	}

	void Draw(HDC hdc, HPEN pen, HBRUSH brush)
	{
		HPEN prev_pen = (HPEN)SelectObject(hdc, pen);
		HBRUSH prev_brush = (HBRUSH)SelectObject(hdc, brush);

		MoveToEx(hdc, pts[0].x, pts[0].y, NULL);
		for(int i = 1; i < (int)pts.size; i++) LineTo(hdc, pts[i].x, pts[i].y);
		
		for(int i = 0; i < (int)pts.size; i++) {
			RECT r = { pts[i].x - GE_POINT_SIZE, pts[i].y - GE_POINT_SIZE, pts[i].x + GE_POINT_SIZE, pts[i].y + GE_POINT_SIZE };
			FrameRect(hdc, &r, brush);
		}

		SelectObject(hdc, prev_pen);
		SelectObject(hdc, prev_brush);
	}

	int GetValue(int pos)
	{
		if(pos < 0) return pts[0].y;
		else if(pos >= GE_CLIENT_WIDTH) return pts[pts.size-1].y;

		int i;
		for(i = 0; i < (int)pts.size-1; i++) {
			if(pos < pts[i+1].x) break;
		}
		if(i == pts.size-1) i--;

		return 255 - pts[i].y + ((255 - pts[i+1].y) - (255 - pts[i].y)) * (pos - pts[i].x) / (pts[i+1].x - pts[i].x);
	}
};

////////////////////////////////////////////////////////////////////////////////
// ϐ錾
////////////////////////////////////////////////////////////////////////////////

HWND g_hwnd_parent;

auls::CWindow g_gewnd;
auls::CPen g_pen;
auls::CBrush g_brush;

CGeGraphData g_graphdata[GE_MAX_GRAPH];

int g_graph_disp;
int g_point_drag;

////////////////////////////////////////////////////////////////////////////////
// OtGfB^֐
////////////////////////////////////////////////////////////////////////////////

void ge_init(void)
{
	for(int i = GE_MAX_GRAPH; i;) {
		i--;
		g_graphdata[i].Init();
	}
}

void ge_SetCaption(void)
{
	SetWindowText(g_hwnd_parent, format("%s [%d/%d]", FILTER_NAME, g_graph_disp+1, GE_MAX_GRAPH));
}

LRESULT CALLBACK ge_wndproc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
	switch(msg) {
		case WM_LBUTTONDOWN:
			if(wp & MK_CONTROL) {
				// CtrlL[A|Cgǉ
				int hit = g_graphdata[g_graph_disp].HitCheck((short)LOWORD(lp), (short)HIWORD(lp));
				bool redraw;
				if(hit == -1) {
					redraw = g_graphdata[g_graph_disp].InsertPoint((short)LOWORD(lp), (short)HIWORD(lp));
				} else {
					redraw = g_graphdata[g_graph_disp].ErasePoint(hit);
				}
				if(redraw) InvalidateRect(hwnd, NULL, TRUE);
				return 0;
			} else {
				// |CghbO
				g_point_drag = g_graphdata[g_graph_disp].HitCheck(LOWORD(lp), HIWORD(lp));
				if(g_point_drag == -1) return 0;
				SetCapture(hwnd);
				return 0;
			}
		case WM_LBUTTONUP:
			if(g_point_drag == -1) return 0;
			ReleaseCapture();
			g_point_drag = -1;
			return 0;
		case WM_MOUSEMOVE:
			if(g_point_drag == -1) return 0;
			g_graphdata[g_graph_disp].MovePoint(g_point_drag, (short)LOWORD(lp), (short)HIWORD(lp));
			InvalidateRect(hwnd, NULL, TRUE);
			return 0;
		case WM_PAINT:
			{
				PAINTSTRUCT ps;
				BeginPaint(hwnd, &ps);
				g_graphdata[g_graph_disp].Draw(ps.hdc, g_pen, g_brush);
				EndPaint(hwnd, &ps);
				SendMessage(g_hwnd_parent, WM_COMMAND, 1234, 0);
				return 0;
			}
		default:
			return DefWindowProc(hwnd, msg, wp, lp);
	}
}

bool ge_createwindow(HWND parent)
{
	WNDCLASSEX wc;
	wc.cbSize = sizeof(wc);
	wc.style  = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = ge_wndproc;
	wc.cbClsExtra  = 0;
	wc.cbWndExtra  = 0;
	wc.hInstance   = NULL;
	wc.hIcon       = (HICON)LoadImage(NULL, MAKEINTRESOURCE(IDI_APPLICATION), IMAGE_ICON, 0, 0, LR_DEFAULTSIZE | LR_SHARED);
	wc.hCursor     = (HCURSOR)LoadImage(NULL, MAKEINTRESOURCE(IDC_ARROW), IMAGE_CURSOR, 0, 0, LR_DEFAULTSIZE | LR_SHARED);
	wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName  = NULL;
	wc.lpszClassName = GE_CLASSNAME;
	wc.hIconSm       = NULL;
	if(!RegisterClassEx(&wc)) return false;

	if(!g_gewnd.Create(GE_CLASSNAME, NULL, WS_CHILD | WS_VISIBLE, 0, 0, 0, 0, 0, parent, NULL, NULL, NULL)) return false;
	auls::SetClientSize(g_gewnd, GE_CLIENT_WIDTH, GE_CLIENT_HEIGHT);
	RECT r;
	GetWindowRect(g_gewnd, &r);
	auls::SetClientSize(parent, r.right - r.left, r.bottom - r.top);
	return true;
}

void ge_SelectPrevGraph(void)
{
	if(g_graph_disp > 0) g_graph_disp--;
	ge_SetCaption();
	InvalidateRect(g_gewnd, NULL, TRUE);
}

void ge_SelectNextGraph(void)
{
	if(g_graph_disp < GE_MAX_GRAPH-1) g_graph_disp++;
	ge_SetCaption();
	InvalidateRect(g_gewnd, NULL, TRUE);
}

void ge_InitGraph(void)
{
	g_graphdata[g_graph_disp].Init();
	InvalidateRect(g_gewnd, NULL, TRUE);
}

////////////////////////////////////////////////////////////////////////////////
// tB^֐
////////////////////////////////////////////////////////////////////////////////

// 
BOOL ge_func_init(FILTER *fp)
{
	g_hwnd_parent = fp->hwnd;
	if(!g_pen.Create(PS_SOLID, 1, GE_GRAPH_COLOR) || !g_brush.CreateSolid(GE_GRAPH_COLOR)) goto failure;
	SetMenu(fp->hwnd, LoadMenu(fp->dll_hinst, MAKEINTRESOURCE(IDR_MENU1)));
	if(!ge_createwindow(fp->hwnd)) goto failure;
	ge_init();
	g_graph_disp = 0;
	g_point_drag = -1;
	ge_SetCaption();
	return TRUE;

failure:
	g_pen.Delete();
	g_brush.Delete();
	MessageBox(GetWindow(fp->hwnd, GW_OWNER), "Ɏs܂B", FILTER_INFORMATION, MB_OK | MB_ICONEXCLAMATION);
	return FALSE;
}

BOOL ge_func_exit(FILTER *fp)
{
	g_gewnd.Destroy();
	return TRUE;
}

BOOL func_WndProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp, void *editp, FILTER *fp)
{
	switch(msg) {
		case WM_KEYDOWN:
			if(wp == VK_LEFT) ge_SelectPrevGraph();
			else if(wp == VK_RIGHT) ge_SelectNextGraph();
			else if(wp == VK_DELETE) { ge_InitGraph(); return TRUE; }
			break;
		case WM_COMMAND:
			switch(LOWORD(wp)) {
				case 1234: return TRUE;
				case ID_PREVGRAPH: ge_SelectPrevGraph(); break;
				case ID_NEXTGRAPH: ge_SelectNextGraph(); break;
				case ID_INITGRAPH: ge_InitGraph(); return TRUE;
				default: break;
			}
	}
	return FALSE;
}

BOOL func_project_load(FILTER *fp, void *editp, void *data, int size)
{
	if(data) {
		memcpy(g_graphdata, data, size);
	}
	return TRUE;
}

BOOL func_project_save(FILTER *fp, void *editp, void *data, int *size)
{
	if(!data) {
		*size = sizeof(g_graphdata);
	} else {
		memcpy(data, g_graphdata, sizeof(g_graphdata));
	}
	return TRUE;
}

FILTER_DLL filter_ge = {
	FILTER_FLAG_DISP_FILTER | FILTER_FLAG_EX_INFORMATION | FILTER_FLAG_ALWAYS_ACTIVE | FILTER_FLAG_WINDOW_SIZE,
	512, 256,
	FILTER_NAME,
	0, NULL, NULL, NULL, NULL,
	0, NULL, NULL,
	NULL,
	ge_func_init,
	ge_func_exit,
	NULL,
	func_WndProc,
	NULL, NULL,
	NULL, 0,
	FILTER_INFORMATION,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	func_project_load,
	func_project_save,
};

extern "C" __declspec(dllexport) FILTER_DLL *GetFilterTable(void)
{
	return &filter_ge;
}




////////////////////////////////////////////////////////////////////////////////
// Lua֐
////////////////////////////////////////////////////////////////////////////////

int l_getvalue(lua_State *L)
{
	// stack
	// 1 : idx / gpOt̔ԍ
	// 2 : rate (0`1 ̒l)

	int idx = lua_tointeger(L, 1);
	double rate = lua_tonumber(L, 2);

	if(idx < 0 || idx >= GE_MAX_GRAPH) lua_pushnumber(L, rate);
	else lua_pushnumber(L, g_graphdata[idx].GetValue((int)(rate * (GE_CLIENT_WIDTH-1))) / 255.0);
	return 1;
}

#define R(name) { #name, l_##name }
luaL_reg table[] = {
	R(getvalue),
	{ NULL, NULL },
};
#undef R

extern "C" __declspec(dllexport) int luaopen_auls_grapheditor(lua_State *L)
{
	luaA_register(L, "auls", "grapheditor", table);
	return 1;
}
